import gzip
import json
import uuid
from binascii import hexlify, unhexlify
from collections import defaultdict
from datetime import datetime, date

from pytz import utc

from tic.mfg.shared.plugins.db.mfg import responses as mfg_responses
from tic.mfg.shared.plugins.db.mfg.responses import BaseClass


class JsonEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, (datetime, date)):
            return o.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        if isinstance(o, mfg_responses.BaseClass):
            return o.__dict__
        return json.JSONEncoder.default(self, o)


def get_new_updates_only(
    reputation_feed_data: list[mfg_responses.ReputationFeedDto],
    all_updated_data: list[(str, int)],
) -> list[mfg_responses.ReputationFeedDto]:
    new_updates: list[mfg_responses.ReputationFeedDto] = list()
    updated_data_dict = dict(all_updated_data)
    for data in reputation_feed_data:
        if data.sha256 in updated_data_dict:
            new_updates.append(data)
    return new_updates


def s3_put_data(
    newly_updated_data: list[mfg_responses.ReputationFeedDto], s3_bucket, s3_client
):
    feed_date = datetime.utcnow().strftime("%Y/%m/%d/%H/%M/%S_%f")[:-4]
    s3_key = f"{feed_date}_{uuid.uuid4()}.gz"
    data = gzip.compress(
        "\n".join(
            str(json.dumps(remove_null(record), cls=JsonEncoder))
            for record in newly_updated_data
        ).encode("utf-8")
    )
    s3_client.put_object(Body=data, Bucket=s3_bucket, Key=s3_key)


def remove_null(value):
    if isinstance(value, BaseClass):
        return remove_null(value.__dict__)
    elif isinstance(value, dict):
        return {key: remove_null(val) for key, val in value.items() if val is not None}
    else:
        return value


def convert_mfg_data(
    mfg_data: list[mfg_responses.MfgReputationData],
) -> dict[str, mfg_responses.ReputationFeedDto]:
    reputation_feed_data: dict[str, mfg_responses.ReputationFeedDto] = dict()
    for data in mfg_data:
        reputation = mfg_responses.Reputation()
        reputation.threat = data.threat
        reputation.threat_locked = data.threat_locked

        dto = mfg_responses.ReputationFeedDto()
        dto.sha256 = data.sha256.lower()
        dto.md5 = data.md5.lower() if data.md5 else data.md5
        dto.sha1 = data.sha1.lower() if data.sha1 else data.sha1
        dto.reputation = reputation
        dto.cb_first_seen = data.date_created
        dto.scan_details = data.scan_details
        dto.sources = data.sources
        dto.mfg = data.mfg
        reputation_feed_data[data.sha256.lower()] = dto
    return reputation_feed_data


def convert_bit9_data(
    bit9data: list[mfg_responses.Bit9ReputationData],
) -> dict[str, mfg_responses.ReputationFeedDto]:
    reputation_feed_data: dict[str, mfg_responses.ReputationFeedDto] = dict()
    for data in bit9data:
        dto = mfg_responses.ReputationFeedDto()
        dto.sha256 = data.sha256.lower()
        dto.md5 = data.md5.lower() if data.md5 else data.md5
        dto.sha1 = data.sha1.lower() if data.sha1 else data.sha1
        dto.tlsh = data.tlsh.lower() if data.tlsh else data.tlsh
        dto.reputation = data.reputation
        dto.file_name = data.file_name
        dto.cb_first_seen = data.cb_first_seen
        dto.sources = data.sources
        dto.signature_info = data.signature_info
        dto.scan_details = data.scan_details
        dto.feed_date = data.feed_date
        reputation_feed_data[data.sha256.lower()] = dto
    return reputation_feed_data


def merge_mfg_bit9_data(
    sha256s: list[str],
    mfg_feed_data: list[mfg_responses.MfgReputationData],
    bit9_feed_data: list[mfg_responses.Bit9ReputationData],
) -> list[mfg_responses.ReputationFeedDto]:
    mfg_data = convert_mfg_data(mfg_feed_data)
    bit9_data = convert_bit9_data(bit9_feed_data)
    feed_date = datetime.utcnow()
    reputation_data: list[mfg_responses.ReputationFeedDto] = list()
    for sha256 in sha256s:
        reputation_item = None
        if sha256.lower() in mfg_data and sha256.lower() in bit9_data:
            mfg = mfg_data[sha256.lower()]
            bit9 = bit9_data[sha256.lower()]
            positive_scanners = list(
                set(
                    (
                        mfg.scan_details.positive_scanners
                        if mfg.scan_details.positive_scanners
                        else list()
                    )
                    + (
                        bit9.scan_details.positive_scanners
                        if bit9.scan_details.positive_scanners
                        else list()
                    )
                )
            )
            scan_details = mfg_responses.ScanDetails()
            scan_details.positive_scanners = positive_scanners
            scan_details.positives = len(positive_scanners)
            scan_details.total_scans = max(
                [
                    mfg.scan_details.total_scans if mfg.scan_details.total_scans else 0,
                    bit9.scan_details.total_scans
                    if bit9.scan_details.total_scans
                    else 0,
                    len(positive_scanners),
                ]
            )
            if not mfg.scan_details.first_scanned_date:
                scan_details.first_scanned_date = bit9.scan_details.first_scanned_date
            elif not bit9.scan_details.first_scanned_date:
                scan_details.first_scanned_date = mfg.scan_details.first_scanned_date
            else:
                scan_details.first_scanned_date = (
                    bit9.scan_details.first_scanned_date
                    if bit9.scan_details.first_scanned_date.astimezone(tz=utc).replace(
                        tzinfo=None
                    )
                    < mfg.scan_details.first_scanned_date
                    else mfg.scan_details.first_scanned_date
                )

            if not mfg.scan_details.last_scanned_date:
                scan_details.last_scanned_date = bit9.scan_details.last_scanned_date
            elif not bit9.scan_details.last_scanned_date:
                scan_details.last_scanned_date = mfg.scan_details.last_scanned_date
            else:
                scan_details.last_scanned_date = (
                    bit9.scan_details.last_scanned_date
                    if bit9.scan_details.last_scanned_date.astimezone(tz=utc).replace(
                        tzinfo=None
                    )
                    > mfg.scan_details.last_scanned_date
                    else mfg.scan_details.last_scanned_date
                )
            if mfg.reputation and mfg.reputation.threat_locked:
                threat = mfg.reputation.threat
            elif bit9.reputation and bit9.reputation.threat_locked:
                threat = bit9.reputation.threat
            else:
                threat = (
                    mfg.reputation.threat
                    if mfg.reputation and mfg.reputation.threat
                    else bit9.reputation and bit9.reputation.threat
                )
            reputation = mfg_responses.Reputation()
            reputation.trust_locked = bit9.reputation and bit9.reputation.trust_locked
            reputation.trust = bit9.reputation and bit9.reputation.trust
            # There are some cases where
            # threat locked will only be locked in bit9 not mfg
            # so if locked in either location use
            # the locked value for threat and threat locked
            reputation.threat_locked = (
                mfg.reputation and mfg.reputation.threat_locked
            ) or (bit9.reputation and bit9.reputation.threat_locked)
            reputation.threat = threat

            reputation_item = mfg_responses.ReputationFeedDto()
            reputation_item.feed_date = bit9.feed_date
            reputation_item.sha256 = sha256.lower()
            reputation_item.md5 = mfg.md5 if mfg.md5 else bit9.md5
            reputation_item.sha1 = mfg.sha1 if mfg.sha1 else bit9.sha1
            reputation_item.tlsh = bit9.tlsh
            reputation_item.reputation = reputation
            reputation_item.file_name = bit9.file_name
            reputation_item.cb_first_seen = (
                bit9.cb_first_seen
                if (
                    bit9.cb_first_seen
                    and bit9.cb_first_seen.astimezone(tz=utc).replace(tzinfo=None)
                )
                < mfg.cb_first_seen
                else mfg.cb_first_seen
            )
            reputation_item.scan_details = scan_details
            reputation_item.sources = list(set(mfg.sources + bit9.sources))
            reputation_item.mfg = mfg.mfg
            reputation_item.signature_info = bit9.signature_info
        elif sha256.lower() in mfg_data:
            reputation_item = mfg_data[sha256.lower()]
        elif sha256.lower() in bit9_data:
            reputation_item = bit9_data[sha256.lower()]
        if reputation_item is None:
            continue
        reputation_item.md5 = (
            reputation_item.md5.lower() if reputation_item.md5 else reputation_item.md5
        )
        reputation_item.sha1 = (
            reputation_item.sha1.lower()
            if reputation_item.sha1
            else reputation_item.sha1
        )
        reputation_item.sha256 = (
            reputation_item.sha256.lower()
            if reputation_item.sha256
            else reputation_item.sha256
        )
        reputation_item.tlsh = reputation_item.tlsh if reputation_item.tlsh else None
        if not reputation_item.feed_date:
            reputation_item.feed_date = feed_date
        if reputation_item:
            reputation_item.minimize()
        reputation_data.append(reputation_item)
    return reputation_data


def get_unique_hashes(
    reputation_feed_data: list[mfg_responses.ReputationFeedDto],
) -> list[(str, int)]:
    reputation_data_list = list()
    for reputation in reputation_feed_data:
        temp_reputation = reputation
        if temp_reputation.sha256:
            temp_reputation.sha256 = reputation.sha256.lower()
        if temp_reputation.sources:
            temp_reputation.sources.sort()
        if temp_reputation.mfg:
            if temp_reputation.mfg.classification_scanners:
                temp_reputation.mfg.classification_scanners.sort()
            if temp_reputation.mfg.tags:
                temp_reputation.mfg.tags.sort()
        if (
            temp_reputation.scan_details
            and temp_reputation.scan_details.positive_scanners
        ):
            temp_reputation.scan_details.positive_scanners = sorted(
                reputation.scan_details.positive_scanners, key=str.casefold
            )
        reputation_data_list.append(temp_reputation)
    # TODO hash value can be different in what C# produced and here
    return [
        (reputation.sha256, hash(json.dumps(reputation, cls=JsonEncoder)))
        for reputation in reputation_data_list
    ]


def convert_to_mfg_response(row: list) -> mfg_responses.MfgInfoResponse:
    response = mfg_responses.MfgInfoResponse()
    response.sha256 = row[0]
    response.sha1 = binary_to_hash(row[1])
    response.md5 = binary_to_hash(row[2])
    response.additional_info = row[3]
    response.platform_num_scanner_votes = row[4]
    response.platform_result = row[5]
    response.platform_description = mfg_responses.ClassificationDescription(row[6])
    response.malware_group_num_scanner_votes = row[7]
    response.malware_group_result = row[8]
    response.malware_group_description = mfg_responses.ClassificationDescription(row[9])
    response.malware_family_num_scanner_votes = row[10]
    response.malware_family_result = row[11]
    response.malware_family_description = mfg_responses.ClassificationDescription(
        row[12]
    )
    response.date_created = row[13]
    response.is_malicious = bool(row[14])
    response.mfg_locked = bool(row[15])
    response.threat = row[16]
    response.threat_locked = bool(row[17])
    return response


def convert_to_sha256_list_response(
    result: dict,
) -> list[mfg_responses.Sha256EnumerableStringResponse]:
    sha256_dict = defaultdict(list)
    for k, v in result:
        sha256_dict[k].append(v)
    sha256_list: list[mfg_responses.Sha256EnumerableStringResponse] = list()
    for k, v in sha256_dict.items():
        response = mfg_responses.Sha256EnumerableStringResponse()
        response.sha256 = k
        response.values = v
        sha256_list.append(response)
    return sha256_list


def convert_to_source_list_response(
    result: dict,
) -> list[mfg_responses.SourcesResponse]:
    source_dict = defaultdict(list)
    for k, v in result:
        source_dict[k].append(v)
    source_list: list[mfg_responses.SourcesResponse] = list()
    for k, v in source_dict.items():
        response = mfg_responses.SourcesResponse()
        response.sha256 = k
        response.sources = v
        source_list.append(response)
    return source_list


def get_sha256_type_array_parameter(hashes: list) -> dict[str, list[str]]:
    return {"sha256": [f"({sha256})" for sha256 in hashes]}


def binary_to_hash(binary_value, type=str):
    if not binary_value:
        return binary_value
    hash_value = hexlify(binary_value)
    if type == str:
        return hash_value.decode("utf-8")
    return hash_value


def hash_to_binary(hash_value):
    if not hash_value:
        return hash_value
    if type(hash_value) == str:
        hash_value = hash_value.encode("utf-8")
    return unhexlify(hash_value)
